"""Parse MoJ XML files"""

import logging
from typing import TypedDict

import lxml.etree as et
import pyproj
import shapely

from .constants import CRS_MAP
from .constants import XML_NAMESPACES as _NS

Point = tuple[float, float]
Curve = tuple[float, float]
Surface = list[list[list[tuple[float, float]]]]

_logger = logging.getLogger(__name__)


class Feature(TypedDict):
    """GeoJSON Feature"""

    type: str
    geometry: dict[str, list]
    properties: dict[str, object]


def _parse_base_properties(root: et._Element) -> dict[str, object]:
    crs_det_elem = root.find("./測地系判別", _NS)
    return {
        "地図名": root.find("./地図名", _NS).text,
        "市区町村コード": root.find("./市区町村コード", _NS).text,
        "市区町村名": root.find("./市区町村名", _NS).text,
        "座標系": root.find("./座標系", _NS).text,
        "測地系判別": crs_det_elem.text if crs_det_elem is not None else None,
    }


def _parse_points(spatial_elem: et._Element) -> dict[str, Point]:
    points: dict[str, Point] = {}
    for point in spatial_elem.iterfind("./zmn:GM_Point", _NS):
        pos = point.find(".//zmn:DirectPosition", _NS)
        x = None
        y = None
        for xy in pos:
            if xy.tag == "{http://www.moj.go.jp/MINJI/tizuzumen}X":
                x = float(xy.text)
            elif xy.tag == "{http://www.moj.go.jp/MINJI/tizuzumen}Y":
                y = float(xy.text)
            else:
                raise ValueError("Unknown tag: {}".format(xy.tag))
        assert x is not None and y is not None
        point_id = point.attrib["id"]
        points[point_id] = (x, y)

    return points


def _parse_curves(
    spatial_elem: et._Element, points: dict[str, Point]
) -> dict[str, Curve]:
    curves: dict[str, Curve] = {}
    for curve in spatial_elem.iterfind("./zmn:GM_Curve", _NS):
        segments = curve.findall("./zmn:GM_Curve.segment", _NS)
        assert len(segments) == 1
        segment = segments[0]

        columns = segment.findall(".//zmn:GM_PointArray.column", _NS)
        assert len(columns) == 2
        column = columns[0]
        assert len(column) == 1
        pos = column[0]
        x = None
        y = None
        if pos.tag == "{http://www.moj.go.jp/MINJI/tizuzumen}GM_Position.indirect":
            ref = pos[0]
            idref = ref.attrib["idref"]
            (x, y) = points[idref]
        elif pos.tag == "{http://www.moj.go.jp/MINJI/tizuzumen}GM_Position.direct":
            for xy in pos:
                if xy.tag == "{http://www.moj.go.jp/MINJI/tizuzumen}X":
                    x = float(xy.text)
                elif xy.tag == "{http://www.moj.go.jp/MINJI/tizuzumen}Y":
                    y = float(xy.text)
                else:
                    raise ValueError("Unknown tag: {}".format(xy.tag))
        else:
            raise ValueError("Unknown tag: {}".format(pos.tag))

        curve_id = curve.attrib["id"]
        assert x is not None and y is not None
        assert curve_id not in curves

        curves[curve_id] = (x, y)

    return curves


def _parse_surfaces(
    spatial_elem: et._Element, curves: dict[str, Curve]
) -> dict[str, Surface]:
    surfaces: dict[str, Surface] = {}
    for surface in spatial_elem.iterfind("./zmn:GM_Surface", _NS):
        assert surface.find(".//zmn:GM_SurfaceBoundary.exterior", _NS) is not None
        polygons = surface.findall("./zmn:GM_Surface.patch/zmn:GM_Polygon", _NS)
        assert len(polygons) == 1
        polygon = polygons[0]

        surface_id = surface.attrib["id"]
        surface_curves: list[tuple[float, float]] = []
        for cc in polygon.iterfind(".//zmn:GM_CompositeCurve.generator", _NS):
            curve_id = cc.attrib["idref"]
            assert curve_id in curves
            assert surface_id not in surface_curves
            surface_curves.append(curves[curve_id])

        assert surface_id not in surfaces
        assert len(surface_curves) > 0, surface_id
        surface_curves.append(surface_curves[0])
        surfaces[surface_id] = [[surface_curves]]

    return surfaces


def _parse_features(
    subject_elem: et._Element, surfaces: dict[str, Surface]
) -> list[Feature]:
    features = []
    for fude in subject_elem.iterfind("./筆", _NS):
        fude_id = fude.attrib["id"]
        properties = {
            "筆ID": fude_id,
            "精度区分": None,
            "丁目名": None,
            "精度区分": None,
            "大字コード": None,
            "丁目コード": None,
            "小字コード": None,
            "予備コード": None,
            "大字名": None,
            "丁目名": None,
            "小字名": None,
            "地番": None,
            "座標値種別": None,
            "筆界未定構成筆": None,
            "地図名": None,
            "市区町村コード": None,
            "市区町村名": None,
            "座標系": None,
            "測地系判別": None,
        }
        geometry = None
        for entry in fude:
            key = entry.tag.split("}")[1]
            if key == "形状":
                coordinates = surfaces[entry.attrib["idref"]]
                geometry = {"type": "MultiPolygon", "coordinates": coordinates}
                rep_point = shapely.MultiPolygon(
                    (p[0], p[1:]) for p in coordinates
                ).point_on_surface()
                properties["代表点経度"] = rep_point.x
                properties["代表点緯度"] = rep_point.y
            else:
                value = entry.text
                properties[key] = value

        # 地番が地区外や別図の場合はスキップする
        chiban = properties.get("地番", "")
        if "地区外" in chiban or "別図" in chiban:
            continue

        features.append(
            {"type": "Feature", "geometry": geometry, "properties": properties}
        )

    return features


# TODO: 仮
def parse_raw(content: bytes) -> list[Feature]:
    """TODO:"""
    doc = et.fromstring(content, None)

    # 基本情報を取得
    base_props = _parse_base_properties(doc)
    source_crs = CRS_MAP[doc.find("./座標系", _NS).text]

    # 任意座標系の場合はスキップ（とりあえず）
    if source_crs is None:
        return []

    _logger.info(f"parsing {base_props['地図名']}...")

    spatial_elem = doc.find("./空間属性", _NS)
    points = _parse_points(spatial_elem)
    curves = _parse_curves(spatial_elem, points)

    # 平面直交座標系を WGS84 に変換する
    if source_crs is not None:
        transformer = pyproj.Transformer.from_crs(
            source_crs, "epsg:4326", always_xy=True
        )
        for curve_id, (x, y) in curves.items():
            curves[curve_id] = transformer.transform(y, x)

    # 小数点以下9ケタに丸める
    for curve_id, (x, y) in curves.items():
        curves[curve_id] = (
            int(x * 1000000000) / 1000000000,
            int(y * 1000000000) / 1000000000,
        )

    surfaces = _parse_surfaces(spatial_elem, curves)

    # Note: 図郭についてはデジタル庁の実装もなんだか変な気がするので一旦扱わない
    # (筆には複数の図郭が結びつく場合があるが、それを考慮できていない)
    # fude_to_zukakus = {}
    # for zk in doc.iterfind(".//図郭", _NS):
    #     zukaku = {
    #         "地図番号": zk.find("地図番号", _NS).text,
    #         "縮尺分母": zk.find("縮尺分母", _NS).text,
    #     }
    #     for fude_ref in zk.iterfind("筆参照", _NS):
    #         fude_id = fude_ref.get("idref")
    #         assert fude_id not in fude_to_zukakus
    #         fude_to_zukakus[fude_id] = zukaku

    subject_elem = doc.find("./主題属性", _NS)
    features = _parse_features(subject_elem, surfaces)

    # XMLのルート要素にある属性情報をFeatureのプロパティに追加する
    for feature in features:
        feature["properties"].update(base_props)

    return features



"""Schema for OGR"""

from collections import OrderedDict
from typing import TypedDict


class _SchemaType(TypedDict):
    properties: OrderedDict[str, str]
    geometry: str


OGR_SCHEMA: _SchemaType = {
    "properties": OrderedDict(
        [
            ("筆ID", "str"),
            ("地図名", "str"),
            ("市区町村コード", "str"),
            ("市区町村名", "str"),
            ("座標系", "str"),
            ("測地系判別", "str"),
            ("大字コード", "str"),
            ("丁目コード", "str"),
            ("小字コード", "str"),
            ("予備コード", "str"),
            ("大字名", "str"),
            ("丁目名", "str"),
            ("小字名", "str"),
            ("地番", "str"),
            ("筆界未定構成筆", "str"),
            ("代表点経度", "float"),
            ("代表点緯度", "float"),
            ("精度区分", "str"),
            ("座標値種別", "str"),
        ]
    ),
    "geometry": "MultiPolygon",
}



"""Command line interface for mojxml"""

import logging
from pathlib import Path

import click

from .process import process_file


@click.command()
@click.argument("dst_file", nargs=1, type=click.Path(dir_okay=False, path_type=Path))
@click.argument(
    "src_files",
    nargs=-1,
    required=True,
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
)
def main(dst_file: Path, src_files: list[Path]) -> None:
    """Convert MoJ XMLs to GeoJSON/GeoPackage/FlatGeobuf/etc.

    DST_FILE: output filename (.geojson, .gpkg, .fgb, etc.)

    SRC_FILES: one or more .xml/.zip files
    """
    # Set up logging
    root_logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter("%(message)s"))
    root_logger.addHandler(handler)
    root_logger.setLevel(logging.INFO)

    # Process files
    process_file(src_paths=src_files, dst_path=dst_file)


if __name__ == "__main__":
    main()



"""Convert .xml/.zip files to OGR format."""

import logging
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Iterable, Optional

import fiona

from .mojzip import MojXMLZipFile
from .parse import parse_raw
from .schema import OGR_SCHEMA

_logger = logging.getLogger(__name__)


def process_raw(
    src_iter: Iterable[bytes], dst_path: Path, driver: Optional[str] = None
) -> None:
    """WIP"""
    with fiona.open(
        dst_path,
        "w",
        driver=driver,
        schema=OGR_SCHEMA,
        crs="EPSG:4326",
    ) as f:
        with ProcessPoolExecutor() as executor:
            count = 0
            for features in executor.map(parse_raw, src_iter):
                if count > 0 and count % 10 == 0:
                    _logger.info(f"{count} files processed")
                for feature in features:
                    f.write(feature)  # type: ignore
                count += 1
            _logger.info(f"{count} files processed")


def _iter_content_xml(src_paths: list[Path]) -> Iterable[bytes]:
    """WIP"""
    for src_path in src_paths:
        src_path = Path(src_path)
        if src_path.suffix == ".xml":
            with open(src_path, "rb") as f:
                yield f.read()
        elif src_path.suffix == ".zip":
            with MojXMLZipFile(src_path) as mzf:
                yield from mzf.iter_xml_contents()
        else:
            raise ValueError(f"Unsupported file type: {src_path.suffix}")


def process_file(src_paths: list[Path], dst_path: Path) -> None:
    """WIP"""
    process_raw(_iter_content_xml(src_paths), dst_path)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mojxml', 'mojxml.mojzip']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'fiona>=1.9.1,<2.0.0',
 'lxml>=4.9.2,<5.0.0',
 'pyproj>=3.4.1,<4.0.0',
 'shapely>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['mojxml2ogr = mojxml.__main__:main']}

setup_kwargs = {
    'name': 'mojxml',
    'version': '0.0.1a6',
    'description': '法務省登記所備付地図データ（地図XML）を読みこむためのPythonライブラリおよび変換用コマンドラインツール',
    'long_description': '# mojxml-py\n\n[![Test](https://github.com/MIERUNE/mojxml-py/actions/workflows/test.yaml/badge.svg)](https://github.com/MIERUNE/mojxml-py/actions/workflows/test.yaml) [![PyPI Package](https://img.shields.io/pypi/v/mojxml?color=%2334D058&label=PyPI%20package)](https://pypi.org/project/mojxml)\n\n法務省登記所備付地図データ（地図XML）を各種GISデータ形式 (GeoJSON, GeoPackage, FlatGeobuf, etc.) に変換するコマンドラインツールです。地図XMLを読み込むためのPythonライブラリとしても使用できます。\n\n## インストール\n\nUbuntu/Debian:\n\n```bash\napt install libgdal-dev\npip3 install mojxml\n```\n\nmacOS (Homebrew):\n\n```bash\nbrew install gdal\npip3 install mojxml\n```\n\n## コマンドラインインタフェース\n\n```\nUsage: mojxml2ogr [OPTIONS] DST_FILE SRC_FILES...\n\n  Convert MoJ XMLs to GeoJSON/GeoPackage/FlatGeobuf/etc.\n\n  DST_FILE: output filename (.geojson, .gpkg, .fgb, etc.)\n\n  SRC_FILES: one or more .xml/.zip files\n```\n\n### 使用例\n\n- 出力形式は拡張子で判断されます。\n- 任意座標系のXMLファイルは無視します（今後オプションを追加）。\n\n```bash\n# XMLファイルをGeoJSONに変換\n❯ mojxml2ogr output.geojson 15222-1107-1553.xml\n\n# 複数のXMLファイルを1つのGeoJSONに変換\n❯ mojxml2ogr output.geojson 15222-1107-1553.xml 15222-1107-1554.xml\n\n# 配布用zipファイルに含まれる全XMLをFlatGeobufに変換\n❯ mojxml2ogr output.fgb 15222-1107.zip\n\n# 3つのzipファイルをまとめて1つのFlatGeobufに変換\n❯ mojxml2ogr output.fgb 01202-4400.zip 01236-4400.zip 01337-4400.zip\n\n# zipファイルを1段階展開して出てくる.zipファイルのうち100個をFlatGeobufに変換\n❯ mojxml2ogr output.fgb 15222-1107-15*.zip\n```\n',
    'author': 'MIERUNE Inc.',
    'author_email': 'info@mierune.co.jp',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/MIERUNE/mojxml-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""法務省登記所備付地図データの多段zip圧縮されたアーカイブを扱う"""

from zipfile import ZipFile


class MojXMLZipFile(ZipFile):
    """法務省登記所備付地図データの多段zip圧縮されたアーカイブを扱う"""

    def iter_xml_contents(self):
        """TODO"""
        for name in self.namelist():
            if name.endswith(".zip"):
                yield self._extract_xml_content(name[:-4])
            elif name.endswith(".xml"):
                yield self.open(name).read()

    def _extract_xml_content(self, internal_name: str) -> bytes:
        """TODO"""
        with self.open(internal_name + ".zip") as f:
            with ZipFile(f) as zf:
                xml_content = zf.open(internal_name + ".xml").read()
                return xml_content



"""定数値"""

from typing import Optional

XML_NAMESPACES: dict[Optional[str], str] = {
    None: "http://www.moj.go.jp/MINJI/tizuxml",
    "zmn": "http://www.moj.go.jp/MINJI/tizuzumen",
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
}

CRS_MAP: dict[str, Optional[str]] = {
    "任意座標系": None,
    "公共座標1系": "epsg:2443",
    "公共座標2系": "epsg:2444",
    "公共座標3系": "epsg:2445",
    "公共座標4系": "epsg:2446",
    "公共座標5系": "epsg:2447",
    "公共座標6系": "epsg:2448",
    "公共座標7系": "epsg:2449",
    "公共座標8系": "epsg:2450",
    "公共座標9系": "epsg:2451",
    "公共座標10系": "epsg:2452",
    "公共座標11系": "epsg:2453",
    "公共座標12系": "epsg:2454",
    "公共座標13系": "epsg:2455",
    "公共座標14系": "epsg:2456",
    "公共座標15系": "epsg:2457",
    "公共座標16系": "epsg:2458",
    "公共座標17系": "epsg:2459",
    "公共座標18系": "epsg:2460",
    "公共座標19系": "epsg:2461",
}
